clc;
clear;

////////////////////////////////////////////////////////////////////////////////
// Compute the optimal control for the cost function
// J = beta/2*integral(u^2) + 1/2*|| sum_i( zout(T, nu(i)) ) - ztarget ||^2
// subject to:
// dz/dt = A(nu(i))*z + B(nu(i))*u
// z(0) = z0
////////////////////////////////////////////////////////////////////////////////

function z=primal(Nt,dt,y0,A,B,u)
    z(:,1)=y0;
    for i=1:length(Nt)-1
        z(:,i+1)=(eye(size(A,1),size(A,1))-dt*A)\(z(:,i)+dt*B*u(i,:));
    end
endfunction

function phi=dual(Nt,dt,phi0,A)
    phi(:,length(Nt))=phi0;
    for i=length(Nt)-1:-1:1
        phi(:,i)=(eye(size(A,1),size(A,1))-dt*A')\(phi(:,i+1));
    end
endfunction


// Vector of parameters
nu = 1:0.5:6;
// Length of vector nu (number of parameters)
M = length(nu);

////////////////////////////////////////////////////////////////////////////////
// System matrices A, B
// zdiff = A*z + B*u
// dz/dt = A*z + B*u
////////////////////////////////////////////////////////////////////////////////
N = 2;
A = eye(N,N);
for i = 1:N
    for j = 1:N
        if j > i
            A(i, j) = 1;
        end
    end
end

A = -A;
B = zeros(N, 1);
B(N) = 1;
////////////////////////////////////////////////////////////////////////////////

// Initial condition
z0 = ones(N, 1);

// Target at t = T
zt = zeros(N, 1); 

// Parameter beta for the cost function
beta = 1e-4;

// Initial time
T0 = 0;
// Final time
T = 1;
// Number of time steps
Nt = 50;
// Time vector
tout = linspace(T0, T, Nt);
dt=tout(2)-tout(1);

// Maximum number of iterations
Nmax = 20000000;

// Control at time k
u = zeros(Nt, 1);
// Control at time k-1
ua = u;
// Control update
Du = zeros(Nt, 1);

// Gradient step
// u = u + d*Du
d = 1;

// Initial err
err = 10;
// Tolerance
tol = 1e-8;
// Iteration counter
iter = 0;

while (err > tol && iter < Nmax)
    // Update iteration counter
    iter = iter + 1;
    
    // Output average: 1/N*sum_i(zout(t, nu(i)))
    zM = zeros(Nt, N);
    
    // Final state
    zend = zeros(M, N);
    
    // Solve primal problem for every parameter in nu
    for j = 1:M
        // Update matrix A
        Am = A + (nu(j) - 1 )*diag(diag(A));
        // Primal problem zdiff = A*z + B*u
        //[tout, zout] = ode45(@(t, z) Am*z + B*interp1(tout, u, t), tout, z0);
        zout=primal(tout,dt,z0,Am,B,u);
        // Save final state for parameter nu(j)
        zend(j, :) = zout(:, $);
        // Update average state
        zM = zM + zout'; 
    end
    // Update average state
    zM = zM/M;
    
    // Initial condition for the adjoint problem
    p0 = - (zM($, :)' - zt);

    // Adjoint variable average: 1/N*sum_i(B*pout(t, nu(i)))
    pM = zeros(Nt, 1);
    for j = 1:M   
        // Update matrix A
        Am = A + (nu(j) - 1 )*diag(diag(A));
        // Solve adjoint problem forward in time
        pout = dual(tout,dt,p0,Am); 
        pM = pM + pout'*B;
    end
    pM = pM/M;
    
    

    // Save previous control
    // u(k-1) = u(k)
    ua = u;
    // Reverse adjoint variable
    //pM = flipud(pM);
    //pM=pM($:-1:1,:);
    
    
     

    // Control update
    Du = beta*u - pM;
    // Update control
    u = u - d*Du;

    for i=1:Nt
        cout_DU2(i)=sqrt(Du(i)^2);
        cout_U2(i)=sqrt(u(i)^2);
    end
    
    err_DU2=sum(cout_DU2.^2)*T/Nt;
    err_U2=sum(cout_U2.^2)*T/Nt;
    
    err=sqrt(err_DU2/err_U2);
                     
        
//    // Control update norm
//    DU2 = integral(@(t) interp1(tout, Du, t).^2, T0, T);
//    U2 = integral(@(t) interp1(tout, u, t).^2, T0, T);
//    
//    err = sqrt(DU2/U2);
//
//    Jmu = 0.5 * (zM(end, :)' - zt)' * (zM(end, :)' - zt);
//    Ju = 0.5 * beta * integral(@(t) interp1(tout, u, t).^2, T0, T);
//
//    J = Jmu + Ju;
//
    printf("Iteration %i - error %g \n", iter, err);
//    
end
//
// Solve state average for the optimal control
// Output average
zM = zeros(Nt, N);
for j = 1:M 
    Am = A + (nu(j) - 1 )*diag(diag(A));
    // Primal problem zdiff = A*z + B*u
    zout=primal(tout,dt,z0,Am,B,u); 
    zM = zM + zout'; 
end
zM = zM/M;

figure(1);
plot(tout',zM(:,1),'r')
plot(tout',zM(:,2),'b')

    figure(3);
    for j = 1:M 
        Am = A + (nu(j) - 1 )*diag(diag(A));
        // Primal problem zdiff = A*z + B*u
        zout=primal(tout,dt,z0,Am,B,u);
        subplot(1, 2, 1)
        plot(tout, zout(1, :))
        subplot(1, 2, 2)
        plot(tout, zout(2, :))
    end 


//
//if N == 2
//    figure(1)
//    plot(tout, zM(:, 1), 'r')
//    hold on
//    plot(tout, zM(:, 2), 'b')
//    legend("zM(1)", "zM(2)")
//    figure(2)
//    plot(tout, u, 'g')
//    legend("u")
//

//end
//



